from keras import ops

from ..core import QLayerBaseMultiInputs

# class QMatmul(QLayerBaseMultiInputs):
#     def build(self, input_shape):
#         assert len(input_shape) == 2, 'QMatmul requires exactly 2 inputs.'
#         super().build(input_shape)

#     def call(self, inputs, training=None):
#         if self.enable_iq:
#             inputs = self.iq(inputs, training=training)
#         inp1, inp2 = inputs
#         return ops.matmul(inp1, inp2)

#     def _compute_ebops(self, *shapes):
#         bits0, bits1 = (iq.bits_(shape) for iq, shape in zip(self.iq, shapes))
#         ebops = ops.sum(ops.matmul(bits0, bits1))
#         return ebops


class QEinsum(QLayerBaseMultiInputs):
    def __init__(self, equation, **kwargs):
        super().__init__(**kwargs)
        self.equation = equation

    def build(self, input_shape):
        super().build(input_shape)
        equation = self.equation
        inp_idxs = [idx.strip() for idx in equation.split('->', 1)[0].split(',')]
        out_idx = equation.split('->', 1)[1].strip()
        inp_ndims = [len(shape) for shape in input_shape]
        assert len(input_shape) == len(inp_idxs), f'Expected {len(inp_idxs)} inputs, but got {len(input_shape)}.'

        wildcard_ndim = None
        available_idxs = set('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ')
        for inp_idx, ndim in zip(inp_idxs, inp_ndims):
            if len(inp_idx) != ndim:
                assert '...' in inp_idx
                if wildcard_ndim is None:
                    wildcard_ndim = ndim - (len(inp_idx) - 3)
                    assert wildcard_ndim >= 0, 'Invalid ellipsis notation: expanded to negative dimensions.'
                assert wildcard_ndim == ndim - (len(inp_idx) - 3), 'Inconsistent ellipsis expansion dimensions across inputs.'
                available_idxs -= set(inp_idx.replace('...', ''))

        if wildcard_ndim is not None:
            rep_indices = list(available_idxs)[:wildcard_ndim]
            rep_indices.sort()
            rep_indices = ''.join(rep_indices)
            for i, inp_idx in enumerate(inp_idxs):
                inp_idxs[i] = inp_idx.replace('...', rep_indices)
            out_idx = out_idx.replace('...', rep_indices)
        self._ebops_equation = ','.join(inp_idxs) + '->'
        self.equation = self._ebops_equation + out_idx

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        return ops.einsum(self.equation, *inputs)

    def _compute_ebops(self, *shapes):
        bitss = [iq.bits_(shape) for iq, shape in zip(self.iq, shapes)]
        ebops = ops.einsum(self._ebops_equation, *bitss)
        return ebops

    def get_config(self):
        config = super().get_config()
        config['equation'] = self.equation
        return config



from keras import ops
from keras.layers import (
    AveragePooling1D,
    AveragePooling2D,
    AveragePooling3D,
    GlobalAveragePooling1D,
    GlobalAveragePooling2D,
    GlobalAveragePooling3D,
    GlobalMaxPooling1D,
    GlobalMaxPooling2D,
    GlobalMaxPooling3D,
    MaxPooling1D,
    MaxPooling2D,
    MaxPooling3D,
)

from .core.base import QLayerBaseSingleInput


class QBasePooling(QLayerBaseSingleInput):
    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        return ops.sum(bw_inp)

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        return super().call(inputs)


class QAveragePooling1D(QBasePooling, AveragePooling1D):
    pass


class QAveragePooling2D(QBasePooling, AveragePooling2D):
    pass


class QAveragePooling3D(QBasePooling, AveragePooling3D):
    pass


class QMaxPooling1D(QBasePooling, MaxPooling1D):
    pass


class QMaxPooling2D(QBasePooling, MaxPooling2D):
    pass


class QMaxPooling3D(QBasePooling, MaxPooling3D):
    pass


class QGlobalAveragePooling1D(QBasePooling, GlobalAveragePooling1D):  # type: ignore
    pass


class QGlobalAveragePooling2D(QBasePooling, GlobalAveragePooling2D):
    pass


class QGlobalAveragePooling3D(QBasePooling, GlobalAveragePooling3D):
    pass


class QGlobalMaxPooling1D(QBasePooling, GlobalMaxPooling1D):
    pass


class QGlobalMaxPooling2D(QBasePooling, GlobalMaxPooling2D):
    pass


class QGlobalMaxPooling3D(QBasePooling, GlobalMaxPooling3D):
    pass


__all__ = [
    'QMaxPooling1D',
    'QMaxPooling2D',
    'QMaxPooling3D',
    'QAveragePooling1D',
    'QAveragePooling2D',
    'QAveragePooling3D',
    'QGlobalMaxPooling1D',
    'QGlobalMaxPooling2D',
    'QGlobalMaxPooling3D',
    'QGlobalAveragePooling1D',
    'QGlobalAveragePooling2D',
    'QGlobalAveragePooling3D',
]



import os
import random
from pathlib import Path

import keras
import numpy as np
import pytest


@pytest.fixture(scope='session', autouse=True, params=[42])
def set_random_seed(request):
    """Set random seeds for reproducibility"""

    seed = request.param
    np.random.seed(seed)
    random.seed(seed)
    backend = keras.backend.backend()
    match backend:
        case 'tensorflow':
            import tensorflow as tf

            tf.random.set_seed(seed)
            for device in tf.config.list_physical_devices('GPU'):
                tf.config.experimental.set_memory_growth(device, True)
        case 'torch':
            import torch

            torch.manual_seed(seed)
        case 'jax':
            pass
        case _:
            raise ValueError(f'Unknown backend: {backend}')


@pytest.fixture(scope='session', autouse=True)
def configure_backend():
    backend = keras.backend.backend()

    match backend:
        case 'tensorflow':
            import tensorflow as tf

            tf.config.threading.set_intra_op_parallelism_threads(1)
        case 'torch':
            import torch

            torch.set_float32_matmul_precision('highest')
        case 'jax':
            pass
        case _:
            raise ValueError(f'Unknown backend: {backend}')


@pytest.fixture(scope='session', autouse=True)
def set_hls4ml_configs():
    """Set default hls4ml configuration"""
    os.environ['HLS4ML_BACKEND'] = 'Vivado'


@pytest.fixture(scope='function')
def temp_directory(request: pytest.FixtureRequest):
    root = Path(os.environ.get('HGQ2_TEST_DIR', '/tmp/hgq2_test'))
    root.mkdir(exist_ok=True)

    test_name = request.node.name
    cls_name = request.cls.__name__ if request.cls else None
    if cls_name is None:
        test_dir = root / test_name
    else:
        test_dir = root / f'{cls_name}.{test_name}'
    test_dir.mkdir(exist_ok=True)
    return str(test_dir)


def pytest_sessionfinish(session, exitstatus):
    """whole test run finishes."""
    root = Path(os.environ.get('HGQ2_TEST_DIR', '/tmp/hgq2_test'))
    # Purge empty directories
    for path in root.glob('*'):
        if path.is_dir() and not any(path.iterdir()):
            path.rmdir()



def str_float_int_or_none(s: str):
    if s == 'None':
        return None
    if s == 'True':
        return True
    if s == 'False':
        return False
    if s.startswith('"') and s.endswith('"'):
        return s[1:-1]
    if s.startswith("'") and s.endswith("'"):
        return s[1:-1]
    if s.isdigit():
        return int(s)
    try:
        return float(s)
    except ValueError:
        pass
    raise ValueError(f'Invalid value: {s}')


def parse_string_to_args(string: str):
    chunks = string.split(',')
    args, kwargs = [], {}
    _pos_arg_ended = False
    for chunk in chunks:
        chunk = chunk.strip()
        if chunk == '':
            raise SyntaxError(f'Invalid syntax: {string}')
        if '=' in chunk:
            _pos_arg_ended = True
            k, v = chunk.split('=')
            k, v = k.strip(), v.strip()
            if k == '' or v == '':
                raise SyntaxError(f'Invalid syntax: {string}')
            kwargs[k] = v
        else:
            assert not _pos_arg_ended, 'Positional argument after keyword argument'
            args.append(chunk)

    args = [str_float_int_or_none(a) for a in args]
    kwargs = {k: str_float_int_or_none(v) for k, v in kwargs.items()}
    return args, kwargs



from ..dataset import Dataset
from .beta_pid import BetaPID
from .beta_scheduler import BetaScheduler, PieceWiseSchedule
from .ebops import FreeEBOPs
from .pareto import ParetoFront
from .pbar import PBar

__all__ = ['BetaPID','BetaScheduler', 'PieceWiseSchedule', 'Dataset', 'FreeEBOPs', 'PBar', 'ParetoFront']



from ..quantizer.config import QuantizerConfig, QuantizerConfigScope, all_quantizer_types, default_configs

__all__ = ['QuantizerConfig', 'QuantizerConfigScope', 'all_quantizer_types', 'default_configs']



from collections.abc import Mapping, Sequence
from typing import TypedDict, overload

from keras.constraints import Constraint
from keras.initializers import Initializer
from keras.regularizers import Regularizer
from keras.saving import deserialize_keras_object, register_keras_serializable

from ..constraints import Min, MinMax
from ..regularizers import MonoL1
from ..utils.misc import numbers
from .internal import (
    BitwidthMapperBase,
    DummyQuantizer,
    FixedPointQuantizerKBI,
    FixedPointQuantizerKIF,
    FloatPointQuantizer,
    TrainableQuantizerBase,
)

default_q_type = {
    'weight': 'kbi',
    'datalane': 'kif',
    'bias': 'kbi',
    'table': 'kbi',
}


class QuantizerConfigBase(TypedDict):
    homogeneous_axis: Sequence[int] | None
    heterogeneous_axis: Sequence[int] | None
    bw_mapper: BitwidthMapperBase | None
    trainable: bool
    is_weight: bool


class KBIConfig(QuantizerConfigBase):
    k0: numbers | bool | Initializer
    b0: numbers | Initializer
    i0: numbers | Initializer
    round_mode: str
    overflow_mode: str
    bc: Constraint | None
    ic: Constraint | None
    br: Regularizer | None
    ir: Regularizer | None
    i_decay_speed: numbers


class KIFConfig(QuantizerConfigBase):
    k0: numbers | bool | Initializer
    i0: numbers | Initializer
    f0: numbers | Initializer
    round_mode: str
    overflow_mode: str
    ic: Constraint | None
    ir: Regularizer | None
    fc: Constraint | None
    fr: Regularizer | None
    i_decay_speed: numbers


class FloatConfig(QuantizerConfigBase):
    m0: numbers | Initializer
    e0: numbers | Initializer
    e00: numbers | Initializer
    mc: Constraint | None
    ec: Constraint | None
    e0c: Constraint | None
    mr: Regularizer | None
    er: Regularizer | None
    e0r: Regularizer | None


kbi_weight_default = KBIConfig(
    k0=True,
    b0=4,
    i0=2,
    round_mode='RND',
    overflow_mode='SAT_SYM',
    bc=MinMax(0, 23),
    ic=None,
    br=MonoL1(1e-8),
    ir=None,
    i_decay_speed=float('inf'),
    homogeneous_axis=(),
    heterogeneous_axis=None,
    bw_mapper=None,
    trainable=True,
    is_weight=True,
)


kbi_datalane_default = KBIConfig(
    k0=True,
    b0=4,
    i0=2,
    round_mode='RND',
    overflow_mode='WRAP',
    bc=MinMax(0, 24),
    ic=None,
    br=MonoL1(1e-8),
    ir=None,
    i_decay_speed=0.01,
    homogeneous_axis=(0,),
    heterogeneous_axis=None,
    bw_mapper=None,
    trainable=True,
    is_weight=False,
)

kif_weight_default = KIFConfig(
    k0=True,
    i0=4,
    f0=2,
    round_mode='RND',
    overflow_mode='SAT_SYM',
    ic=MinMax(-23, 23),
    ir=MonoL1(1e-8),
    fc=MinMax(-24, 24),
    fr=MonoL1(1e-8),
    i_decay_speed=float('inf'),
    homogeneous_axis=(),
    heterogeneous_axis=None,
    bw_mapper=None,
    trainable=True,
    is_weight=True,
)


kif_datalane_default = KIFConfig(
    k0=True,
    i0=4,
    f0=2,
    round_mode='RND',
    overflow_mode='WRAP',
    ic=MinMax(-23, 23),
    ir=MonoL1(1e-8),
    fc=MinMax(-24, 24),
    fr=MonoL1(1e-8),
    i_decay_speed=0.01,
    homogeneous_axis=(0,),
    heterogeneous_axis=None,
    bw_mapper=None,
    trainable=True,
    is_weight=False,
)

float_weight_default = FloatConfig(
    m0=2,
    e0=4,
    e00=0,
    mc=MinMax(-1, 8),
    ec=MinMax(0, 4),
    e0c=MinMax(-16, 16),
    mr=MonoL1(1e-8),
    er=MonoL1(1e-8),
    e0r=None,
    homogeneous_axis=(),
    heterogeneous_axis=None,
    bw_mapper=None,
    trainable=True,
    is_weight=True,
)


float_datalane_default = FloatConfig(
    m0=2,
    e0=4,
    e00=0,
    mc=MinMax(-1, 8),
    ec=MinMax(0, 4),
    e0c=MinMax(-16, 16),
    mr=MonoL1(1e-8),
    er=MonoL1(1e-8),
    e0r=None,
    homogeneous_axis=None,
    heterogeneous_axis=(),
    bw_mapper=None,
    trainable=True,
    is_weight=False,
)

kbi_table_default = kbi_weight_default.copy()
kif_table_default = kif_weight_default.copy()
float_table_default = float_weight_default.copy()

kbi_table_default['homogeneous_axis'] = None
kbi_table_default['heterogeneous_axis'] = ()
kbi_table_default['is_weight'] = False
kif_table_default['homogeneous_axis'] = None
kif_table_default['heterogeneous_axis'] = ()
kif_table_default['is_weight'] = False
float_table_default['homogeneous_axis'] = None
float_table_default['heterogeneous_axis'] = ()
float_table_default['is_weight'] = False

default_configs: dict[tuple[str, str], KIFConfig | KBIConfig | FloatConfig] = {
    ('kbi', 'weight'): kbi_weight_default,
    ('kbi', 'bias'): kbi_weight_default.copy(),
    ('kbi', 'table'): kbi_table_default,
    ('kbi', 'datalane'): kbi_datalane_default,
    ('kif', 'weight'): kif_weight_default,
    ('kif', 'bias'): kif_weight_default.copy(),
    ('kif', 'table'): kif_table_default,
    ('kif', 'datalane'): kif_datalane_default,
    ('float', 'weight'): float_weight_default,
    ('float', 'bias'): float_weight_default.copy(),
    ('float', 'table'): float_table_default,
    ('float', 'datalane'): float_datalane_default,
}

all_quantizer_keys = {k for v in default_configs.values() for k in v.keys()} | {'q_type', 'place', 'scaler', 'qnoise_factor'}


def all_quantizer_types():
    return {k[0] for k in default_configs.keys()}


def all_places():
    return {k[1] for k in default_configs.keys()}


@register_keras_serializable(package='hgq')
class QuantizerConfig(Mapping):
    @overload
    def __init__(
        self,
        q_type: str = 'kbi',
        place: str = 'datalane',
        *,
        k0: numbers | bool | Initializer = True,
        b0: numbers | Initializer = 4,
        i0: numbers | Initializer = 2,
        round_mode: str = 'RND',
        overflow_mode: str = 'WRAP',
        bc: Constraint | None = MinMax(0, 12),
        ic: Constraint | None = None,
        br: Regularizer | None = None,
        ir: Regularizer | None = None,
        i_decay_speed: numbers = -1,
        homogeneous_axis: Sequence[int] | None = None,
        heterogeneous_axis: Sequence[int] | None = None,
        bw_mapper: BitwidthMapperBase | None = None,
        scaler: numbers | None = None,
        qnoise_factor: float | None = None,
        **kwargs,
    ) -> None:
        """Fixed point quantizer config with KBI parametrization.

        Parameters
        ----------
        q_type : str
            The type of the quantizer. 'kbi' for this implementation.
        place : str
            Where the quantizer is expected to be place. Only affects default config. One of 'weight', 'datalane', 'bias', and 'table'.
        k0 : numbers | bool | Initializer, optional
            If the quantizer allows negative values, by default True
        b0 : numbers | Initializer, optional
            The initial value of the number of bits (excl. sign), by default 4
        i0 : numbers | Initializer, optional
            The initial value of the number of integer bits (excl. sign), by default 2
        round_mode : str, optional
            Rounding mode. One of 'RND', 'RND_CONV', 'TRN', 'S_RND', 'S_RND_CONV', by default 'RND'
        overflow_mode : str, optional
            Overflow mode. One of 'WRAP', 'SAT', 'SAT_SYM', by default 'WRAP'
        bc : Constraint | None, optional
            Constraint for the number of bits, by default MinMax(0, 12)
        ic : Constraint | None, optional
            Constraint for the number of integer bits, by default None
        br : Regularizer | None, optional
            Regularizer for the number of bits, by default None
        ir : Regularizer | None, optional
            Regularizer for the number of integer bits, by default None
        i_decay_speed : numbers, optional
            The decay speed of the integer. Only used if `round_mode` is 'WRAP', by default -1
        homogeneous_axis : Sequence[int] | None, optional
            The axes that are quantized homogeneously. Mutually exclusive with `heterogeneous_axis`. Only used if `bw_mapper` is not set, by default None
        heterogeneous_axis : Sequence[int] | None, optional
            The axes that are quantized heterogeneously. Mutually exclusive with `homogeneous_axis`. Only used if `bw_mapper` is not set, by default None
        bw_mapper : BitwidthMapperBase | None, optional
            The bitwidth mapper to be used. Must be a subclass of `BitwidthMapperBase`. If None, the default bitwidth mapper is used with `homogeneous_axis` and `heterogeneous_axis` as arguments, by default None
        scaler : numbers | None, optional
            The scaling factor to be used. If None, no scaling is applied, by default None
        qnoise_factor : float | None, optional
            The fraction of quantization strength. If None, treat as full quantization noise, by default None
        """
        ...

    @overload
    def __init__(
        self,
        q_type: str = 'kif',
        place: str = 'datalane',
        *,
        k0: numbers | bool | Initializer = True,
        i0: numbers | Initializer = 4,
        f0: numbers | Initializer = 2,
        round_mode: str = 'RND',
        overflow_mode: str = 'WRAP',
        ic: Constraint | None = MinMax(-12, 12),
        ir: Regularizer | None = None,
        fc: Constraint | None = MinMax(-10, 10),
        fr: Regularizer | None = None,
        i_decay_speed: numbers = 0.01,
        homogeneous_axis: Sequence[int] | None = (0,),
        heterogeneous_axis: Sequence[int] | None = None,
        bw_mapper: BitwidthMapperBase | None = None,
        scaler: numbers | None = None,
        qnoise_factor: float | None = None,
        **kwargs,
    ) -> None:
        """Fixed point quantizer config with KIF parametrization.

        Parameters
        ----------
        q_type : str
            The type of the quantizer. 'kif' for this implementation.
        place : str
            Where the quantizer is expected to be place. Only affects default config. One of 'weight', 'datalane', 'bias', and 'table'.
        k0 : numbers | bool | Initializer, optional
            If the quantizer allows negative values, by default True
        i0 : numbers | Initializer, optional
            The initial value of the number of integer bits (excl. sign), by default 4
        f0 : numbers | Initializer, optional
            The initial value of the number of fraction bits, by default 2
        round_mode : str, optional
            Rounding mode. One of 'RND', 'RND_CONV', 'TRN', 'S_RND', 'S_RND_CONV', by default 'RND'
        overflow_mode : str, optional
            Overflow mode. One of 'WRAP', 'SAT', 'SAT_SYM', by default 'SAT'
        ic : Constraint | None, optional
            Constraint for the number of integer bits, by default MinMax(-12, 12)
        ir : Regularizer | None, optional
            Regularizer for the number of integer bits, by default None
        fc : Constraint | None, optional
            Constraint for the number of fraction bits, by default MinMax(-12, 12)
        fr : Regularizer | None, optional
            Regularizer for the number of fraction bits, by default None
        i_decay_speed : numbers, optional
            The decay speed of the integer. Only used if `round_mode` is 'WRAP', by default 0.01
        homogeneous_axis : Sequence[int] | None, optional
            The axes that are quantized homogeneously. Mutually exclusive with `heterogeneous_axis`. Only used if `bw_mapper` is not set, by default (0,)
        heterogeneous_axis : Sequence[int] | None, optional
            The axes that are quantized heterogeneously. Mutually exclusive with `homogeneous_axis`. Only used if `bw_mapper` is not set, by default None
        bw_mapper : BitwidthMapperBase | None, optional
            The bitwidth mapper to be used. Must be a subclass of `BitwidthMapperBase`. If None, the default bitwidth mapper is used with `homogeneous_axis` and `heterogeneous_axis` as arguments, by default None
        scaler : numbers | None, optional
            The scaling factor to be used. If None, no scaling is applied, by default None
        qnoise_factor : float | None, optional
            The fraction of quantization strength. If None, treat as full quantization noise, by default None
        """
        ...

    @overload
    def __init__(
        self,
        q_type: str = 'float',
        place: str = 'datalane',
        *,
        m0: numbers | Initializer = 2,
        e0: numbers | Initializer = 1,
        e00: numbers | Initializer = 0,
        mc: Constraint | None = Min(-1),
        ec: Constraint | None = MinMax(0, 4),
        e0c: Constraint | None = MinMax(-8, 8),
        mr: Regularizer | None = None,
        er: Regularizer | None = None,
        e0r: Regularizer | None = None,
        homogeneous_axis: Sequence[int] | None = (),
        heterogeneous_axis: Sequence[int] | None = None,
        bw_mapper: BitwidthMapperBase | None = None,
        scaler: numbers | None = None,
        qnoise_factor: float | None = None,
        **kwargs,
    ) -> None:
        """Floating point quantizer config.

        Parameters
        ----------
        q_type : str
            The type of the quantizer. 'float' for this implementation.
        place : str
            Where the quantizer is expected to be place. Only affects default config. One of 'weight', 'datalane', 'bias', and 'table'.
        m0 : numbers | Initializer, optional
            The initial value of the number of mantissa bits, by default 2
        e0 : numbers | Initializer, optional
            The initial value of the number of exponent bits, by default 1
        e00 : numbers | Initializer, optional
            The initial value of the number of exponent bits for the first axis, by default 0
        mc : Constraint | None, optional
            Constraint for the number of mantissa bits, by default Min(-1)
        ec : Constraint | None, optional
            Constraint for the number of exponent bits, by default MinMax(0, 4)
        e0c : Constraint | None, optional
            Constraint for the number of exponent bits for the first axis, by default MinMax(-8, 8)
        mr : Regularizer | None, optional
            Regularizer for the number of mantissa bits, by default None
        er : Regularizer | None, optional
            Regularizer for the number of exponent bits, by default None
        e0r : Regularizer | None, optional
            Regularizer for the number of exponent bits for the first axis, by default None
        homogeneous_axis : Sequence[int] | None, optional
            The axes that are quantized homogeneously. Mutually exclusive with `heterogeneous_axis`. Only used if `bw_mapper` is not set, by default ()
        heterogeneous_axis : Sequence[int] | None, optional
            The axes that are quantized heterogeneously. Mutually exclusive with `homogeneous_axis`. Only used if `bw_mapper` is not set, by default None
        bw_mapper : BitwidthMapperBase | None, optional
            The bitwidth mapper to be used. Must be a subclass of `BitwidthMapperBase`. If None, the default bitwidth mapper is used with `homogeneous_axis` and `heterogeneous_axis` as arguments, by default None
        scaler : numbers | None, optional
            The scaling factor to be used. If None, no scaling is applied, by default None
        qnoise_factor : float | None, optional
            The fraction of quantization strength. If None, treat as full quantization noise, by default None
        """
        ...

    def __init__(
        self,
        q_type: str = 'default',
        place: str = 'datalane',
        scaler: numbers | None = None,
        qnoise_factor: float | None = None,
        **kwargs,
    ) -> None:
        """Universal quantizer config. The type of the quantizer is specified by the `type` argument.

        Parameters
        ----------
        q_type : str
            The type of the quantizer. One of 'kbi', 'kif', 'float', 'default'. If 'default', the default quantizer type is used, by default 'kbi'. Can be overridden by the `default_q_type` argument of `QuantizerConfigScope`.
        place : str, optional
            The default config to be loaded of the quantizer. One of 'weight', 'datalane', by default 'weight'
        scaler : numbers | None, optional
            The scaling factor to be used. If None, no scaling is applied, by default None
        qnoise_factor : float | None, optional
            The fraction of quantization strength. If None, treat as full quantization noise, by default None

        **kwargs : Specific parameters for different quantizer types.
        """

        place = place.lower()
        q_type = q_type.lower()
        if q_type == 'default':
            q_type = default_q_type[place]
        if q_type != 'dummy':
            assert (q_type, place) in default_configs, f'Default config for ({q_type}, {place}) not found.'
        self.place = place
        self.q_type = q_type

        self.scaler = None
        self.qnoise_factor = None
        if scaler is not None:
            assert scaler != 0, 'scaler must not be 0.'
            self.scaler = float(scaler)
        if qnoise_factor is not None:
            assert 0 <= qnoise_factor <= 1, 'qnoise_factor must be between 0 and 1.'
            self.qnoise_factor = float(qnoise_factor) if qnoise_factor is not None else None

        if q_type == 'dummy':  # Special case for dummy quantizer
            self.config = {}
            return

        assert kwargs.get('homogeneous_axis') is None or kwargs.get('heterogeneous_axis') is None, (
            'homogeneous_axis and heterogeneous_axis are mutually exclusive. Set only one of them.'
        )

        if kwargs.get('homogeneous_axis') is not None:
            kwargs['heterogeneous_axis'] = None
        if kwargs.get('heterogeneous_axis') is not None:
            kwargs['homogeneous_axis'] = None

        config = default_configs.get((q_type, place))
        assert config is not None, f'Default config for ({q_type}, {place}) not found.'
        self.config = config.copy()

        if self.config is not None:
            for k, v in kwargs.items():
                if k not in self.config:
                    raise ValueError(f'{k} is not a valid parameter for {q_type} quantizer config.')
                self.config[k] = v

    def __getitem__(self, key):
        return self.config[key]

    def __iter__(self):
        return iter(self.config)

    def __len__(self) -> int:
        return len(self.config)

    def __bool__(self):
        return True

    def get_config(self):
        return {
            'q_type': self.q_type,
            'place': self.place,
            'scaler': self.scaler,
            'qnoise_factor': self.qnoise_factor,
            **self.config,
        }

    @classmethod
    def from_config(cls, config):
        return cls(**deserialize_keras_object(config))

    def get_quantizer(self) -> TrainableQuantizerBase:
        match self.q_type:
            case 'float':
                return FloatPointQuantizer(**self)  # type: ignore
            case 'kif':
                return FixedPointQuantizerKIF(**self)  # type: ignore
            case 'kbi':
                return FixedPointQuantizerKBI(**self)  # type: ignore
            case 'dummy':
                return DummyQuantizer()
            case _:
                raise ValueError(f'Unknown quantizer type: {self.q_type}')


class QuantizerConfigScope:
    def __init__(
        self,
        q_type: str | Sequence[str] | set[str] = 'all',
        place: str | Sequence[str] | set[str] = 'all',
        default_q_type=None,
        **kwargs,
    ):
        """Override default quantizer config within a context.

        Parameters
        ----------
        q_type : str
            The type of the quantizers.
        place : str
            The location of the quantizers.
        default_q_type : str, optional
            The default quantizer type to be used. If None, the default quantizer type is not changed. One of 'kbi', 'kif', 'float', by default None
        """

        if q_type == 'all':
            q_type = all_quantizer_types()
        if place == 'all':
            place = all_places()

        q_type = (q_type,) if isinstance(q_type, str) else q_type
        place = (place,) if isinstance(place, str) else place
        q_type = {_q_type.lower() for _q_type in q_type}
        place = {_place.lower() for _place in place}

        for _q_type in q_type:
            for _place in place:
                assert (_q_type, _place) in default_configs, f'Default config for ({_q_type}, {_place}) not found.'

        assert kwargs.get('homogeneous_axis') is None or kwargs.get('heterogeneous_axis') is None, (
            'homogeneous_axis and heterogeneous_axis are mutually exclusive. Set only one of them.'
        )

        if kwargs.get('homogeneous_axis') is not None:
            kwargs['heterogeneous_axis'] = None
        if kwargs.get('heterogeneous_axis') is not None:
            kwargs['homogeneous_axis'] = None

        i, f, b = kwargs.get('i0', None), kwargs.get('f0', None), kwargs.get('b0', None)
        if sum((i is not None, f is not None, b is not None)) == 2:
            if i is None:
                kwargs['i0'] = b - f
            if f is None:
                kwargs['f0'] = b - i
            if b is None:
                kwargs['b0'] = i + f

        for k in kwargs:
            if k not in all_quantizer_keys:
                raise ValueError(f'{k} is not a valid parameter for any known quantizer configs.')

        self.q_types = q_type
        self.places = place
        self.kwargs = kwargs
        self.default_q_type = default_q_type
        self._tmp_storage = {}
        self.original_default_q_type = None

    def __enter__(self):
        for (q_type, place), default_conf in default_configs.items():
            if q_type in self.q_types and place in self.places:
                self._tmp_storage[(q_type, place)] = default_conf.copy()
                for k, v in self.kwargs.items():
                    if k in default_conf:
                        default_conf[k] = v
        if self.default_q_type is not None:
            self.original_default_q_type = default_q_type.copy()
            for place in self.places:
                default_q_type[place] = self.default_q_type

    def __exit__(self, exc_type, exc_value, traceback):
        if self.original_default_q_type is not None:
            default_q_type.clear()
            default_q_type.update(self.original_default_q_type)
            self.original_default_q_type = None

        for q_type, place in self._tmp_storage:
            default_configs[(q_type, place)].update(self._tmp_storage[(q_type, place)])
        self._tmp_storage.clear()

    def override(self):
        """Override the default quantizer config."""
        self.__enter__()
        self._tmp_storage.clear()



from collections.abc import Sequence

import keras
from keras import ops
from keras.utils import PyDataset
from numpy.typing import ArrayLike
from tqdm import tqdm

from .dataset import Dataset


def _reset_minmax(layer: keras.Layer):
    if hasattr(layer, '_i_decay_speed'):
        # WRAP-like overflow mode
        shape, dtype = layer._i.shape, layer._i.dtype
        layer._i.assign(keras.ops.full(shape, -1e9, dtype=dtype))
        shape, dtype = layer._k.shape, layer._k.dtype
        layer._k.assign(keras.ops.zeros(shape, dtype=dtype))
    for sublayer in layer._layers:
        _reset_minmax(sublayer)


class TrainingFlagWrapper:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other

    def __bool__(self):
        return self.value is True


def trace_minmax(
    model: keras.Model,
    data: ArrayLike | Sequence[ArrayLike] | PyDataset,
    reset=True,
    batch_size=1024,
    verbose: int | bool = 0,
    return_results=False,
):
    n_outputs = len(model.outputs)

    if not isinstance(data, PyDataset):
        data = Dataset(data, batch_size=batch_size, device='none')

    if reset:
        _reset_minmax(model)
    record: dict[str, int] = {}

    results = []
    use_pbar = verbose is True or verbose > 1
    n_batch = len(data)  # type: ignore
    n_outputs = len(model.outputs)

    with tqdm(total=n_batch, leave=False, disable=not use_pbar) as pbar:  # type: ignore
        for i in range(n_batch):
            r = model(data[i][0], training=TrainingFlagWrapper('tracing'))
            if return_results:
                results.append(r)
            pbar.update(1)

    if verbose:
        record = {}
        for layer in model.layers:
            if getattr(layer, 'enable_ebops', False):
                record[layer.name] = int(layer.ebops)  # type: ignore
        width = max(max(map(len, record.keys())), 5)
        for k, v in record.items():
            print(f'{k:{width}}: {v}')
        print(f'Total: {sum(record.values())}')

    if return_results:
        if n_outputs == 1:
            return ops.stop_gradient(ops.concatenate(results))
        return tuple(ops.stop_gradient(ops.concatenate([r[i] for r in results])) for i in range(n_outputs))



from math import log2, prod

import numpy as np
import pytest
from keras import layers, ops

from hgq.config import QuantizerConfigScope
from hgq.layers import QLayerBase
from hgq.layers.ops import QAdd, QAveragePow2, QDot, QEinsum, QMaximum, QMeanPow2, QMinimum, QMultiply, QSubtract, QSum
from tests.base import LayerTestBase


class MergeOpsBase(LayerTestBase):
    layer_cls = QLayerBase
    keras_layer_cls = layers.Layer

    @pytest.fixture(params=[2])
    def input_shapes(self, request):
        shape = (2, 4, 2)
        return [shape] * request.param

    def test_behavior(self, input_data, layer_kwargs, *args, **kwargs):
        with QuantizerConfigScope(default_q_type='dummy'):
            hgq_layer = self.layer_cls(**layer_kwargs)
            keras_layer = self.keras_layer_cls(**layer_kwargs)

        r_hgq: np.ndarray = ops.convert_to_numpy(hgq_layer(input_data))  # type: ignore
        r_keras: np.ndarray = ops.convert_to_numpy(keras_layer(input_data))  # type: ignore
        np.testing.assert_allclose(r_hgq, r_keras)


class TestQAdd(MergeOpsBase):
    layer_cls = QAdd
    keras_layer_cls = layers.Add


class TestQSubtract(MergeOpsBase):
    layer_cls = QSubtract
    keras_layer_cls = layers.Subtract


class TestQDot(MergeOpsBase):
    layer_cls = QDot
    keras_layer_cls = layers.Dot
    # hls4ml qdot supports only 1d vectors

    @pytest.fixture
    def layer_kwargs(self):
        return {'axes': -1}

    @pytest.fixture(params=[2])
    def input_shapes(self, request):
        shape = (8,)
        return [shape] * request.param

    @pytest.fixture(params=[True])
    def use_parallel_io(self, request) -> bool:
        return request.param


class TestQEinsum(MergeOpsBase):
    layer_cls = QEinsum

    @pytest.fixture
    def input_shapes(self, request):
        return ((2, 4, 3), (2, 3, 4))

    @pytest.fixture(params=['Babc,Bacd->Babd', '...abd,...adb->...'])
    def equation(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, equation):
        return {'equation': equation}

    def test_behavior(self, input_data, layer_kwargs):
        with QuantizerConfigScope(default_q_type='dummy'):
            hgq_layer = self.layer_cls(**layer_kwargs)

        r_hgq: np.ndarray = ops.convert_to_numpy(hgq_layer(input_data))  # type: ignore
        r_keras: np.ndarray = ops.convert_to_numpy(ops.einsum(layer_kwargs['equation'], *input_data))  # type: ignore
        np.testing.assert_allclose(r_hgq, r_keras)

    @pytest.fixture(params=[True])
    def use_parallel_io(self, request) -> bool:
        return request.param


class TestQMaximum(MergeOpsBase):
    layer_cls = QMaximum
    keras_layer_cls = layers.Maximum


class TestQMinimum(MergeOpsBase):
    layer_cls = QMinimum
    keras_layer_cls = layers.Minimum


class TestQMultiply(MergeOpsBase):
    layer_cls = QMultiply
    keras_layer_cls = layers.Multiply


class TestQAveragePow2(MergeOpsBase):
    layer_cls = QAveragePow2

    @pytest.fixture(params=[2])
    def input_shapes(self, request):
        shape = (2, 4, 3)
        return [shape] * request.param

    def test_behavior(self, input_data):
        with QuantizerConfigScope(default_q_type='dummy'):
            hgq_layer = self.layer_cls()

        s = 2.0 ** round(log2(1.0 / len(input_data)))

        r_hgq: np.ndarray = ops.convert_to_numpy(hgq_layer(input_data))  # type: ignore
        r_keras: np.ndarray = ops.convert_to_numpy(ops.sum(input_data, axis=0) * s)  # type: ignore
        np.testing.assert_allclose(r_hgq, r_keras)


# class TestQMatmul(MergeOpsBase):
#     layer_cls = QMatmul

#     @pytest.fixture(params=[((3, 5), (5, 6)), ((3, 4, 5), (3, 5, 6))])
#     def input_shapes(self, request):
#         return request.param

#     def test_behavior(self, input_data):
#         with QuantizerConfigScope(default_q_type='dummy'):
#             hgq_layer = self.layer_cls()

#         r_hgq: np.ndarray = ops.convert_to_numpy(hgq_layer(input_data))  # type: ignore
#         r_keras: np.ndarray = ops.convert_to_numpy(ops.matmul(*input_data))  # type: ignore

#         np.testing.assert_allclose(r_hgq, r_keras)


class TestSum(LayerTestBase):
    layer_cls = QSum
    hls4ml_not_supported = True

    @pytest.fixture(params=[(8,), (2, 3, 4)])
    def input_shapes(self, request):
        return request.param

    @pytest.fixture
    def axes(self, input_shapes):
        if len(input_shapes) == 1:
            return 1
        else:
            return (-2, -1)

    @pytest.fixture
    def layer_kwargs(self, axes):
        return {'axes': axes}

    @pytest.fixture
    def scale(self, input_shapes, axes):
        return 1.0

    def test_behavior(self, input_data, layer_kwargs, scale):
        with QuantizerConfigScope(default_q_type='dummy'):
            hgq_layer = self.layer_cls(**layer_kwargs)

        r_hgq: np.ndarray = ops.convert_to_numpy(hgq_layer(input_data))  # type: ignore
        r_keras: np.ndarray = ops.convert_to_numpy(ops.sum(input_data, axis=layer_kwargs['axes']) * scale)  # type: ignore
        np.testing.assert_allclose(r_hgq, r_keras)


class TestMeanPow2(TestSum):
    layer_cls = QMeanPow2
    hls4ml_not_supported = True

    @pytest.fixture
    def scale(self, input_shapes: tuple[int, ...], axes):
        if isinstance(axes, int):
            scale = 1.0 / input_shapes[axes % (len(input_shapes) + 1) - 1]
        else:
            scale = 1.0 / prod(input_shapes[ax % (len(input_shapes) + 1) - 1] for ax in axes)
        return 2.0 ** round(log2(scale))



"""S-QUARK
===================================================
Scalable Quantization-Aware Realtime Keras
---------------------------------------------------

The S-QUARK library provides a set of tools to quantize neural networks meant to be deployed on edge devices, mainly FPGAs with Keras. The library is designed to be scalable, allowing for the construction of fully-quantized models suitable for deployment with minimal effort.

Provides
--------
- Scalable quantization-aware training
- Drop-in replacement quantized keras layers
- Support for various quantization schemes
- Trainable weights and quantization bitwidths
- TensorFlow/JAX/PyTorch backend support with Keras v3

Library Structure
-----------------
The library is organized in a keras-like structure, with the following modules:

- `config`: Configuration settings for the layers and quantizers
- `layers`: Quantized keras layers
- `quantizer`: Quantizer wrappers and internal quantizers
- `utils`: Utility functions and classes, and some useful sugars
- `constraints`: Custom constraints for quantization-aware training
- `regularizers`: Custom regularizers for quantization-aware training

Compatibility
-------------
- hls4ml: WIP
- QKeras: Never, as it is built on Keras v2. However, this library comes with a QKeras-like compatibility API. Refer to the `qkeras` module (top level, not under this) for more information.

"""

from . import config, layers, quantizer, utils
from ._version import __version__  # noqa: F401

__all__ = ['config', 'layers', 'quantizer', 'utils']



import keras
import pytest

from hgq.layers import QBatchNormDense, QDense, QEinsumDense, QEinsumDenseBatchnorm
from tests.base import LayerTestBase


class TestDense(LayerTestBase):
    da4ml_not_supported = False
    layer_cls = QDense

    @pytest.fixture(params=[8])  # Test different output sizes
    def units(self, request):
        return request.param

    @pytest.fixture(params=[None, 'relu'])  # Test with and without activation
    def activation(self, request):
        return request.param

    @pytest.fixture(params=[(8, 8), (12,)])
    def input_shapes(self, request):
        return request.param

    @pytest.fixture
    def layer_kwargs(self, units, activation):
        return {'units': units, 'activation': activation}

    def test_da4ml_conversion(self, model: keras.Model, input_data, overflow_mode: str, temp_directory: str):
        super()._test_da4ml_conversion(
            model=model,
            input_data=input_data,
            overflow_mode=overflow_mode,
            temp_directory=temp_directory,
        )


class TestBatchNormDense(TestDense):
    layer_cls = QBatchNormDense


class TestEinsumDense(LayerTestBase):
    layer_cls = QEinsumDense

    @pytest.fixture(params=['Bij,ijk->Bik', 'Babc,bcd->Babd'])
    def equation(self, request):
        return request.param

    @pytest.fixture
    def input_shapes(self, equation: str):
        if equation == 'Bij,ijk->Bik':
            return (3, 4)
        return (3, 4, 5)

    @pytest.fixture
    def output_shape(self, equation: str):
        if equation == 'Bij,ijk->Bik':
            return (3, 6)
        return (3, 4, 2)

    @pytest.fixture(params=[False, True])
    def bias_axes(self, equation: str, request) -> str | None:
        if not request.param:
            return None
        if equation == 'Bij,ijk->Bik':
            return 'ik'
        return 'ad'

    @pytest.fixture
    def layer_kwargs(self, equation: str, output_shape: tuple[int, ...], bias_axes: str | None):
        return {'equation': equation, 'output_shape': output_shape, 'bias_axes': bias_axes, 'activation': None}

    @pytest.fixture(params=[True])
    def use_parallel_io(self, request) -> bool:
        return request.param

    def test_da4ml_conversion(self, model: keras.Model, input_data, overflow_mode: str, temp_directory: str):
        super()._test_da4ml_conversion(
            model=model,
            input_data=input_data,
            overflow_mode=overflow_mode,
            temp_directory=temp_directory,
        )


class TestEinsumDenseBatchnorm(TestEinsumDense):
    layer_cls = QEinsumDenseBatchnorm

    @pytest.fixture()
    def bias_axes(self, equation: str, request):
        if equation == 'Bij,ijk->Bik':
            return 'ik'
        return 'bd'



from .layer import LayerConfigScope, global_config
from .quantizer import QuantizerConfig, QuantizerConfigScope, default_configs

__all__ = ['LayerConfigScope', 'QuantizerConfigScope', 'QuantizerConfig', 'global_config', 'default_configs']



import inspect

from hgq import layers
from hgq.config import QuantizerConfigScope

from ..initializers import QInitializer as QInitializer  # TODO: use qkeras initializers when called from here
from ..quantizers import get_quantizer

kw_map = {
    'kq_conf': ('kernel_quantizer', 'kq'),
    'bq_conf': ('bias_quantizer', 'bq'),
    'oq_conf': ('output_quantizer', 'oq'),
    'iq_conf': ('input_quantizer', 'iq'),
    'exp_iq_conf': ('exp_input_quantizer', 'exp_iq'),
    'exp_oq_conf': ('exp_output_quantizer', 'exp_oq'),
    'inv_iq_conf': ('inv_input_quantizer', 'inv_iq'),
    'inv_oq_conf': ('inv_output_quantizer', 'inv_oq'),
}

kw_map_inv = {vv: k for k, v in kw_map.items() for vv in v}


def qkeras_layer_wrap(cls: type):
    # base_cls = cls.__bases__[0]
    original_init = cls.__init__
    signature = inspect.signature(original_init)
    params = signature.parameters
    new_params = []
    for v in params.values():
        if v.name in kw_map:
            new_params.append(v.replace(name=kw_map[v.name][0]))
        else:
            new_params.append(v)
    new_signature = signature.replace(parameters=new_params)

    original_init = cls.__init__

    def __init__(self, *args, **kwargs):
        for k, v in list(kwargs.items()):
            if k not in kw_map_inv:
                continue
            new_k = kw_map_inv[k]
            assert new_k not in kwargs, f'Duplicate key {new_k}.'
            del kwargs[k]
            v = v if not isinstance(v, str) else get_quantizer(v)
            kwargs[new_k] = v

        # Disable EBOPS; only explicitly enabled quantizers will be used
        kwargs['enable_ebops'] = False
        kwargs['enable_iq'] = kwargs.get('iq_conf') is not None
        kwargs['enable_oq'] = kwargs.get('oq_conf') is not None
        with QuantizerConfigScope(default_q_type='dummy'):
            return original_init(self, *args, **kwargs)

    __init__.__signature__ = new_signature  # type: ignore
    cls.__init__ = __init__
    return cls


for name, obj in layers.__dict__.items():
    if not isinstance(obj, type):
        continue
    if issubclass(obj, layers.QLayerBase):
        globals()[name] = qkeras_layer_wrap(obj)



from collections.abc import Callable

from keras import ops
from keras.layers import Activation
from keras.src import backend

from ..quantizer import QuantizerConfig
from .core.base import QLayerBaseSingleInput


def _large_negative_number(dtype):
    """Return a Large negative number based on dtype."""
    if backend.standardize_dtype(dtype) == 'float16':
        return -3e4
    return -1e9


class QUnaryFunctionLUT(Activation, QLayerBaseSingleInput):
    def __init__(
        self,
        activation: Callable | str,
        iq_conf: QuantizerConfig | None = None,
        oq_conf: QuantizerConfig | None = None,
        enable_oq=True,
        enable_iq=True,
        allow_heterogeneous_input: bool = False,
        allow_heterogeneous_table: bool = False,
        override_oq_k0_to_0: bool = False,
        **kwargs,
    ):
        act_name = activation.__name__ if isinstance(activation, Callable) else activation
        assert act_name not in ('softmax', 'log_softmax'), f'activation {act_name} is not unary'
        assert not allow_heterogeneous_table, 'No hls4ml support; remove this check if you know what you are doing.'

        self._allow_heterogeneous_table = allow_heterogeneous_table
        self._allow_heterogeneous_input = allow_heterogeneous_input

        if enable_oq:
            oq_conf = oq_conf or QuantizerConfig('default', 'table')
            if override_oq_k0_to_0:
                if 'k0' in oq_conf.config:
                    oq_conf.config['k0'] = False
            if not allow_heterogeneous_table:
                oq_conf.config['homogeneous_axis'] = None
                oq_conf.config['heterogeneous_axis'] = ()
        else:
            self._enable_ebops = False

        if enable_iq and not allow_heterogeneous_input:
            iq_conf = iq_conf or QuantizerConfig('default', 'datalane')
            iq_conf.config['homogeneous_axis'] = None
            iq_conf.config['heterogeneous_axis'] = ()

        super().__init__(
            activation=activation, iq_conf=iq_conf, oq_conf=oq_conf, enable_oq=enable_oq, enable_iq=enable_iq, **kwargs
        )
        self.built = False

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        x = super().call(inputs)
        if self.enable_oq:
            x = self.oq(x, training=training)
        return x

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_out = self.oq.bits_(shape)
        # TODO: more realistic cost for lookup tables
        return ops.sum((2.0**bw_inp) * bw_out) * 1e-4  # type: ignore

    def get_config(self):
        config = super().get_config()
        config['allow_heterogeneous_table'] = self._allow_heterogeneous_table
        config['allow_heterogeneous_input'] = self._allow_heterogeneous_input
        return config


# class QPositiveUnaryFunctionLUT(QUnaryFunctionLUT):
#     def __init__(
#         self,
#         activation: Callable | str,
#         iq_conf: QuantizerConfig | None = None,
#         oq_conf: QuantizerConfig | None = None,
#         allow_heterogeneous_table: bool = False,
#         **kwargs,
#     ):
#         assert kwargs.pop(
#             'enable_out_quantizer', True
#         ), 'enable_out_quantizer must be True for QPositiveUnaryFunctionLUT, if set.'
#         super().__init__(
#             activation=activation,
#             iq_conf=iq_conf,
#             oq_conf=oq_conf,
#             enable_oq=True,
#             allow_heterogeneous_table=allow_heterogeneous_table,
#             override_oq_k0_to_0=True,
#             **kwargs,
#         )

#     def call(self, inputs, training=None):
#         x = super().call(inputs, training=training)
#         eps = self.oq.epsilon_(ops.shape(inputs))
#         return ops.maximum(x, eps)



from collections.abc import Sequence
from math import log2, prod

from keras import ops

from ...quantizer.config import QuantizerConfig
from ...utils.misc import warn_no_synth
from ..core.base import QLayerBaseSingleInput


class QSum(QLayerBaseSingleInput):
    def __init__(
        self,
        iq_conf: QuantizerConfig | None = None,
        axes: int | Sequence[int] = -1,
        scale: float = 1.0,
        keepdims: bool = False,
        **kwargs,
    ):
        super().__init__(iq_conf=iq_conf, **kwargs)
        assert kwargs.get('axis', None) is None, f'Use "axes" instead of "axis" in {self.__class__.__name__} ({self.name}).'
        self.axes = tuple(axes) if isinstance(axes, Sequence) else (axes,)
        assert log2(scale).is_integer(), 'Scale must be a power of 2.'
        self._scale = scale
        self._keepdims = keepdims

    def build(self, input_shape):
        super().build(input_shape)
        axes = sorted(i if i >= 0 else i + len(input_shape) for i in self.axes)
        self.axes = tuple(axes)
        cond = not all(i1 - i0 == 1 for i0, i1 in zip(axes[:-1], axes[1:]))
        warn_no_synth(cond, 'Summing non-adjacent axes may not be supported by the synthesis tool.')

    @property
    def scale(self):
        return self._scale

    @property
    def keepdims(self):
        return self._keepdims

    def _compute_ebops(self, shape):
        bits = self.iq.bits_(shape)
        ebops = ops.sum(bits) - ops.sum(ops.min(bits, axis=self.axes))  # type: ignore
        ebops = ebops * 0.65  # TODO: better ebops cost model for accumulators
        return ebops

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        r = ops.sum(inputs, axis=self.axes, keepdims=self.keepdims) * self.scale  # type: ignore
        return r

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'axes': self.axes,
                'scale': self.scale,
                'keepdims': self.keepdims,
            }
        )
        return config


class QMeanPow2(QSum):
    def __init__(
        self,
        iq_conf: QuantizerConfig | None = None,
        axes: int | Sequence[int] = -1,
        keepdims: bool = False,
        **kwargs,
    ):
        super().__init__(iq_conf=iq_conf, axes=axes, keepdims=keepdims, **kwargs)

    def build(self, input_shape):
        super().build(input_shape)
        scale = 1.0 / prod([input_shape[i] for i in self.axes])
        self._scale = 2.0 ** round(log2(scale))

    def get_config(self):
        config = super().get_config()
        config.pop('scale')
        return config



"""S-QUARK/QKeras
QKeras compatibility layer for S-QUARK. This API is for compatibility purpose only, and only exposes a minimal subset of the S-QUARK's functionality. It is not recommended for new projects.

If you are starting a new project, please use the S-QUARK API directly for more flexibility and control.
"""

from .quantizers import *



from keras import ops
from keras.src.layers.convolutional.base_conv import BaseConv

from ..quantizer import Quantizer
from ..quantizer.config import QuantizerConfig
from ..utils.misc import gather_vars_to_kwargs, warn_no_synth
from .core.base import QLayerBaseSingleInput


class QBaseConv(QLayerBaseSingleInput, BaseConv):
    def __init__(
        self,
        rank,
        filters,
        kernel_size,
        strides=1,
        padding='valid',
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        parallelization_factor=-1,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self|kq_conf|bq_conf|parallelization_factor')
        super().__init__(**kwargs)

        kq_conf = kq_conf or QuantizerConfig('default', 'weight')
        self._kq = Quantizer(kq_conf, name=f'{self.name}_kq')
        bq_conf = bq_conf or QuantizerConfig('default', 'bias')
        self._bq = Quantizer(bq_conf, name=f'{self.name}_bq') if use_bias else None
        self.parallelization_factor = parallelization_factor

        dilation_is_1 = dilation_rate == 1 if isinstance(dilation_rate, int) else all(d == 1 for d in dilation_rate)  # type: ignore
        warn_no_synth(not dilation_is_1, 'Dilation rate is not supported by hls4ml.')

    @property
    def kq(self):
        return self._kq

    @property
    def bq(self):
        return self._bq

    def build(self, input_shape):
        super().build(input_shape)
        self.kq.build(ops.shape(self.kernel))
        if self.use_bias:
            assert self.bq is not None
            self.bq.build(ops.shape(self.bias))

    def call(self, inputs, training=None):
        if self.enable_iq:
            inputs = self.iq(inputs, training=training)
        outputs = self.convolution_op(
            inputs,
            self.qkernel,
        )
        if self.use_bias:
            if self.data_format == 'channels_last':
                bias_shape = (1,) * (self.rank + 1) + (self.filters,)
            else:
                bias_shape = (1, self.filters) + (1,) * self.rank
            qbias = ops.reshape(self.qbias, bias_shape)
            outputs += qbias  # type: ignore

        if self.activation is not None:
            return self.activation(outputs)
        return outputs

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_ker = self.kq.bits_(ops.shape(self.kernel))
        if self.parallelization_factor < 0:
            ebops = ops.sum(self.convolution_op(bw_inp, bw_ker))
        else:
            reduce_axis_kernel = tuple(range(0, self.rank + 1))
            if self.data_format == 'channels_last':
                reduce_axis_input = reduce_axis_kernel
            else:
                reduce_axis_input = (0,) + tuple(range(2, self.rank + 2))

            bw_inp = ops.max(bw_inp, axis=reduce_axis_input)  # Keep only maximum per channel
            reduce_axis_kernel = tuple(range(0, self.rank))
            bw_ker = ops.sum(bw_ker, axis=reduce_axis_kernel)  # Keep only sum per channel
            ebops = ops.sum(bw_inp[:, None] * bw_ker)

        if self.bq is not None:
            size = ops.cast(ops.prod(shape), self.dtype)
            bw_bias = self.bq.bits_(ops.shape(self.bias))
            ebops = ebops + ops.mean(bw_bias) * size  # type: ignore

        return ebops

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                'kq_conf': self.kq.config,
                'bq_conf': self.bq.config if self.bq is not None else None,
                'parallelization_factor': self.parallelization_factor,
            }
        )
        return config

    @property
    def qkernel(self):
        return self.kq(self._kernel)

    @property
    def qbias(self):
        if self.bias is None:
            return None
        assert self.bq is not None
        return self.bq(self.bias)


class QConv1D(QBaseConv):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=1,
        padding='valid',
        data_format=None,
        dilation_rate=1,
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self')
        super().__init__(rank=1, **kwargs)

    def _compute_causal_padding(self):
        left_pad = self.dilation_rate[0] * (self.kernel_size[0] - 1)
        if self.data_format == 'channels_last':
            causal_padding = [[0, 0], [left_pad, 0], [0, 0]]
        else:
            causal_padding = [[0, 0], [0, 0], [left_pad, 0]]
        return causal_padding

    def call(self, inputs, training=None):
        padding = self.padding
        if self.padding == 'causal':
            # Apply causal padding to inputs.
            inputs = ops.pad(inputs, self._compute_causal_padding())
            padding = 'valid'

        qinputs = self.iq(inputs, training=training)
        qkernel = self.kq(self._kernel, training=training)

        outputs = ops.conv(
            qinputs,
            qkernel,
            strides=list(self.strides),  # type: ignore
            padding=padding,
            dilation_rate=self.dilation_rate,  # type: ignore
            data_format=self.data_format,
        )

        if self.use_bias:
            if self.data_format == 'channels_last':
                bias_shape = (1,) * (self.rank + 1) + (self.filters,)
            else:
                bias_shape = (1, self.filters) + (1,) * self.rank
            qbias = self.bq(self.bias, training=training)  # type: ignore
            qbias = ops.reshape(qbias, bias_shape)
            outputs += qbias  # type: ignore
        if self.activation is not None:
            return self.activation(outputs)
        return outputs


class QConv2D(QBaseConv):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1),
        padding='valid',
        data_format=None,
        dilation_rate=(1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self')
        super().__init__(rank=2, **kwargs)


class QConv3D(QBaseConv):
    def __init__(
        self,
        filters,
        kernel_size,
        strides=(1, 1),
        padding='valid',
        data_format=None,
        dilation_rate=(1, 1),
        groups=1,
        activation=None,
        use_bias=True,
        kernel_initializer='glorot_uniform',
        bias_initializer='zeros',
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        kq_conf: None | QuantizerConfig = None,
        iq_conf: None | QuantizerConfig = None,
        bq_conf: None | QuantizerConfig = None,
        **kwargs,
    ):
        kwargs = gather_vars_to_kwargs('self')
        warn_no_synth(True, 'QConv3D is not supported by hls4ml.')
        super().__init__(rank=3, **kwargs)



from keras import ops
from keras.saving import deserialize_keras_object, register_keras_serializable, serialize_keras_object
from keras.src.layers.input_spec import InputSpec
from keras.src.layers.rnn.simple_rnn import RNN, SimpleRNN, SimpleRNNCell

from ...config import QuantizerConfig
from ...quantizer import Quantizer
from ..core.base import QLayerBaseSingleInput


class QSimpleRNNCell(QLayerBaseSingleInput, SimpleRNNCell):
    __no_wrap_call__ = True

    @property
    def kq(self):
        "Kernel Quantizer"
        return self._kq

    @property
    def rkq(self):
        "Recurrent Kernel Quantizer"
        return self._rkq

    @property
    def iq(self):
        "Input Quantizer"
        return self._iq

    @property
    def sq(self):
        "State Quantizer"
        return self._sq

    @property
    def bq(self):
        "Bias Quantizer"
        return self._bq

    @property
    def qkernel(self):
        return self.kq(self.kernel)

    @property
    def qrecurrent_kernel(self):
        return self.rkq(self.recurrent_kernel)

    @property
    def qbias(self):
        return self.bq(self.bias)

    def __init__(
        self,
        units,
        activation='tanh',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        recurrent_initializer='orthogonal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        seed=None,
        iq_conf: QuantizerConfig | None = None,
        sq_conf: QuantizerConfig | None = None,
        kq_conf: QuantizerConfig | None = None,
        rkq_conf: QuantizerConfig | None = None,
        bq_conf: QuantizerConfig | None = None,
        **kwargs,
    ):
        iq_conf = iq_conf or QuantizerConfig(place='datalane')
        sq_conf = sq_conf or QuantizerConfig(place='datalane')
        kq_conf = kq_conf or QuantizerConfig(place='weight')
        rkq_conf = rkq_conf or QuantizerConfig(place='weight')
        bq_conf = bq_conf or QuantizerConfig(place='bias')
        # self._iq = Quantizer(iq_conf, name='iq')
        self._sq = Quantizer(sq_conf, name='sq')
        self._kq = Quantizer(kq_conf, name='kq')
        self._rkq = Quantizer(rkq_conf, name='rkq')
        self._bq = Quantizer(bq_conf, name='bq')

        super().__init__(
            units=units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            iq_conf=iq_conf,
            seed=seed,
            **kwargs,
        )

    def build(self, input_shape):
        self._sq.build((None, self.units))
        self._kq.build((input_shape[-1], self.units))
        self._rkq.build((self.units, self.units))
        self._bq.build((self.units,))
        super().build(input_shape)

    def call(self, sequence, states, training=False):
        prev_output = states[0] if isinstance(states, (list, tuple)) else states
        dp_mask = self.get_dropout_mask(sequence)
        rec_dp_mask = self.get_recurrent_dropout_mask(prev_output)

        if training and dp_mask is not None:
            sequence = sequence * dp_mask
        if training and rec_dp_mask is not None:
            prev_output = prev_output * rec_dp_mask

        qkernel = self.kq(self.kernel, training=training)
        qrecurrent_kernel = self.rkq(self.recurrent_kernel, training=training)
        qstate = self.sq(prev_output, training=training)
        qsequence = self.iq(sequence, training=training)

        h = ops.matmul(qsequence, qkernel)
        if self.bias is not None:
            h += self.bq(self.bias, training=training)

        output = h + ops.matmul(qstate, qrecurrent_kernel)  # type: ignore
        if self.activation is not None:
            output = self.activation(output)

        new_state = [output] if isinstance(states, (list, tuple)) else output
        return output, new_state

    def get_config(self):
        conf = super().get_config()
        conf.update(
            {
                'iq_conf': self._iq.config,
                'sq_conf': self._sq.config,
                'kq_conf': self._kq.config,
                'rkq_conf': self._rkq.config,
                'bq_conf': self._bq.config,
            }
        )
        return conf

    def _compute_ebops(self, shape):
        bw_inp = self.iq.bits_(shape)
        bw_ker = self.kq.bits_(ops.shape(self.kernel))
        ebops1 = ops.sum(ops.matmul(bw_inp, bw_ker))
        bw_state = self.sq.bits_((1, self.units))
        bw_rker = self.rkq.bits_(ops.shape(self.recurrent_kernel))
        ebops2 = ops.sum(ops.matmul(bw_state, bw_rker))
        ebops = ebops1 + ebops2  # type: ignore
        if self.bq is not None:
            bw_bias = self.bq.bits_(ops.shape(self.bias))
            size = ops.cast(ops.prod(shape), self.dtype)
            ebops = ebops + ops.mean(bw_bias) * size  # type: ignore

        return ebops


@register_keras_serializable(package='hgq')
class QSimpleRNN(SimpleRNN):
    """Fully-connected RNN where the output is to be fed back as the new input.

    Args:
        units: Positive integer, dimensionality of the output space.
        activation: Activation function to use.
            Default: hyperbolic tangent (`tanh`).
            If you pass None, no activation is applied
            (ie. "linear" activation: `a(x) = x`).
        use_bias: Boolean, (default `True`), whether the layer uses
            a bias vector.
        kernel_initializer: Initializer for the `kernel` weights matrix,
            used for the linear transformation of the inputs. Default:
            `"glorot_uniform"`.
        recurrent_initializer: Initializer for the `recurrent_kernel`
            weights matrix, used for the linear transformation of the recurrent
            state.  Default: `"orthogonal"`.
        bias_initializer: Initializer for the bias vector. Default: `"zeros"`.
        kernel_regularizer: Regularizer function applied to the `kernel` weights
            matrix. Default: `None`.
        recurrent_regularizer: Regularizer function applied to the
            `recurrent_kernel` weights matrix. Default: `None`.
        bias_regularizer: Regularizer function applied to the bias vector.
            Default: `None`.
        activity_regularizer: Regularizer function applied to the output of the
            layer (its "activation"). Default: `None`.
        kernel_constraint: Constraint function applied to the `kernel` weights
            matrix. Default: `None`.
        recurrent_constraint: Constraint function applied to the
            `recurrent_kernel` weights matrix.  Default: `None`.
        bias_constraint: Constraint function applied to the bias vector.
            Default: `None`.
        dropout: Float between 0 and 1.
            Fraction of the units to drop for the linear transformation
            of the inputs. Default: 0.
        recurrent_dropout: Float between 0 and 1.
            Fraction of the units to drop for the linear transformation of the
            recurrent state. Default: 0.
        return_sequences: Boolean. Whether to return the last output
            in the output sequence, or the full sequence. Default: `False`.
        return_state: Boolean. Whether to return the last state
            in addition to the output. Default: `False`.
        go_backwards: Boolean (default: `False`).
            If `True`, process the input sequence backwards and return the
            reversed sequence.
        stateful: Boolean (default: `False`). If `True`, the last state
            for each sample at index i in a batch will be used as initial
            state for the sample of index i in the following batch.
        unroll: Boolean (default: `False`).
            If `True`, the network will be unrolled,
            else a symbolic loop will be used.
            Unrolling can speed-up a RNN,
            although it tends to be more memory-intensive.
            Unrolling is only suitable for short sequences.
        iq_conf: QuantizerConfig, optional
            Input Quantizer configuration
        sq_conf: QuantizerConfig, optional
            State Quantizer configuration
        kq_conf: QuantizerConfig, optional
            Kernel Quantizer configuration
        rkq_conf: QuantizerConfig, optional
            Recurrent Kernel Quantizer configuration
        bq_conf: QuantizerConfig, optional
            Bias Quantizer configuration

    Call arguments:
        sequence: A 3D tensor, with shape `[batch, timesteps, feature]`.
        mask: Binary tensor of shape `[batch, timesteps]` indicating whether
            a given timestep should be masked. An individual `True` entry
            indicates that the corresponding timestep should be utilized,
            while a `False` entry indicates that the corresponding timestep
            should be ignored.
        training: Python boolean indicating whether the layer should behave in
            training mode or in inference mode.
            This argument is passed to the cell when calling it.
            This is only relevant if `dropout` or `recurrent_dropout` is used.
        initial_state: List of initial state tensors to be passed to the first
            call of the cell.

    Example:

    ```python
    inputs = np.random.random((32, 10, 8))
    simple_rnn = keras.layers.SimpleRNN(4)
    output = simple_rnn(inputs)  # The output has shape `(32, 4)`.
    simple_rnn = keras.layers.SimpleRNN(4, return_sequences=True, return_state=True)
    # whole_sequence_output has shape `(32, 10, 4)`.
    # final_state has shape `(32, 4)`.
    whole_sequence_output, final_state = simple_rnn(inputs)
    ```
    """

    def __init__(
        self,
        units,
        activation='tanh',
        use_bias=True,
        kernel_initializer='glorot_uniform',
        recurrent_initializer='orthogonal',
        bias_initializer='zeros',
        kernel_regularizer=None,
        recurrent_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        recurrent_constraint=None,
        bias_constraint=None,
        dropout=0.0,
        recurrent_dropout=0.0,
        return_sequences=False,
        return_state=False,
        go_backwards=False,
        stateful=False,
        unroll=False,
        seed=None,
        iq_conf: QuantizerConfig | None = None,
        sq_conf: QuantizerConfig | None = None,
        kq_conf: QuantizerConfig | None = None,
        rkq_conf: QuantizerConfig | None = None,
        bq_conf: QuantizerConfig | None = None,
        parallelization_factor=-1,
        **kwargs,
    ):
        cell = QSimpleRNNCell(
            units,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            recurrent_initializer=recurrent_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            recurrent_regularizer=recurrent_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            recurrent_constraint=recurrent_constraint,
            bias_constraint=bias_constraint,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            seed=seed,
            dtype=kwargs.get('dtype', None),
            trainable=kwargs.get('trainable', True),
            name='simple_rnn_cell',
            iq_conf=iq_conf,
            sq_conf=sq_conf,
            kq_conf=kq_conf,
            rkq_conf=rkq_conf,
            bq_conf=bq_conf,
        )
        RNN.__init__(
            self,
            cell,
            return_sequences=return_sequences,
            return_state=return_state,
            go_backwards=go_backwards,
            stateful=stateful,
            unroll=unroll,
            **kwargs,
        )
        self.input_spec = [InputSpec(ndim=3)]
        self.parallelization_factor = parallelization_factor

    def get_config(self):  # type: ignore
        conf = super().get_config()
        conf.update(
            {
                'iq_conf': self.cell._iq.config,
                'sq_conf': self.cell._sq.config,
                'kq_conf': self.cell._kq.config,
                'rkq_conf': self.cell._rkq.config,
                'bq_conf': self.cell._bq.config,
                'parallelization_factor': self.parallelization_factor,
            }
        )
        return serialize_keras_object(conf)

    @classmethod
    def from_config(cls, config):
        config = deserialize_keras_object(config)
        pass
        return super().from_config(config)

    def build(self, sequences_shape, initial_state_shape=None):
        seq_len = sequences_shape[1]
        if self.parallelization_factor == -1:
            self.parallelization_factor = seq_len
        return super().build(sequences_shape, initial_state_shape)

    def call(self, sequences, initial_state=None, mask=None, training=False):
        ebops = self.cell._compute_ebops((1,) + ops.shape(sequences[0])[1:]) * self.parallelization_factor
        self.cell._ebops.assign(ops.cast(ebops, self.cell._ebops.dtype))  # type: ignore
        self.add_loss(ebops * self.cell.beta)
        return super().call(sequences, mask=mask, training=training, initial_state=initial_state)



from .accum import QMeanPow2, QSum
from .einsum import QEinsum
from .merge import QAdd, QAveragePow2, QDot, QMaximum, QMinimum, QMultiply, QSubtract

__all__ = [
    'QMeanPow2',
    'QSum',
    'QAdd',
    'QAveragePow2',
    'QDot',
    'QEinsum',
    'QMaximum',
    'QMinimum',
    'QMultiply',
    'QSubtract',
]
